using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Collections;

namespace kmeans
{
    class KMeans
    {
        class Row
        {
            public List<int> m_files = new List<int>(0);
            public List<short> m_data = new List<short>(0);
            public List<int> m_pos = new List<int>(0);
            public double norm = 0.0;
        }

        class AssignedCategory
        {
            public int m_fileNum;
            public int m_category;
        }

        List<Row> m_rows = new List<Row>(0);
        List<Row> m_centroid = new List<Row>(0);
        List<AssignedCategory> m_categorizedFiles = new List<AssignedCategory>(0);
 
        private float getCosine(Row r1, Row r2)
        {
            double product = 0.0f;
            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    product += r1.m_data[index1] * r2.m_data[index2];
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }
            product /= r1.norm;
            product /= r2.norm;
            double min = r1.m_files.Count;
            if (min > r2.m_files.Count) min = r2.m_files.Count;
            product /= min;
            return (float)(product);
        }

        private void getMaxElement(ref int nRow, ref int nCol, ref float fMax, float[,] fMatrix, int nSize)
        {
            nRow = 0;
            nCol = 0;
            fMax = fMatrix[nRow, nCol];
            for (int i = 0; i < nSize; ++i)
            {
                for (int j = i + 1; j < nSize; ++j)
                {
                    if (fMax < fMatrix[i, j])
                    {
                        fMax = fMatrix[i, j];
                        nRow = i;
                        nCol = j;
                    }
                }
            }
        }

        private Row mergeTwoRows(Row r1, Row r2)
        {
            Row merged = new Row();

            for (int i = 0; i < r1.m_files.Count; ++i)
            {
                merged.m_files.Add(r1.m_files[i]);
            }
            for (int i = 0; i < r2.m_files.Count; ++i)
            {
                merged.m_files.Add(r2.m_files[i]);
            }

            int index1 = 0;
            int index2 = 0;
            while (true)
            {
                if (r1.m_pos[index1] == r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    short sum = r1.m_data[index1];
                    sum += r2.m_data[index2];
                    merged.m_data.Add(sum);
                    ++index1;
                    ++index2;
                }
                else if (r1.m_pos[index1] > r2.m_pos[index2])
                {
                    merged.m_pos.Add(r2.m_pos[index2]);
                    merged.m_data.Add(r2.m_data[index2]);
                    ++index2;
                }
                else if (r1.m_pos[index1] < r2.m_pos[index2])
                {
                    merged.m_pos.Add(r1.m_pos[index1]);
                    merged.m_data.Add(r1.m_data[index1]);
                    ++index1;
                }
                if (index1 >= r1.m_pos.Count) break;
                if (index2 >= r2.m_pos.Count) break;
            }

            for (int i = index1; i < r1.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r1.m_pos[i]);
                merged.m_data.Add(r1.m_data[i]);
            }

            for (int i = index2; i < r2.m_pos.Count; ++i)
            {
                merged.m_pos.Add(r2.m_pos[i]);
                merged.m_data.Add(r2.m_data[i]);
            }

            merged.norm = 0.0;
            foreach (short s in merged.m_data)
            {
                merged.norm += s * s;
            }
            merged.norm = Math.Sqrt(merged.norm);

            return merged;
        }

        private bool computeInitialCentroids(string fileName, int nCategories)
        {
            if (nCategories < 2)
            {
                Console.WriteLine("Wrong number of categories");
                return false;
            }
            if (!File.Exists(fileName))
            {
                return false;
            }

            FileInfo fi = new FileInfo(fileName);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);
            ArrayList list = new ArrayList();
            list.Clear();

            //read data
            int nR = 0;
            int nC = 0;
            short sum = 0;
            int current_row = 0;
            m_rows.Add(new Row());
            m_rows[current_row].m_files.Add(current_row);
            using (FileStream stream = new FileStream(fileName, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nRnew = reader.ReadInt32();
                        nC = reader.ReadInt32();
                        sum = reader.ReadInt16();
                        if (nRnew != nR)
                        {
                            m_rows.Add(new Row());
                            ++current_row;
                            m_rows[current_row].m_files.Add(current_row);
                        }
                        nR = nRnew;
                        m_rows[current_row].m_pos.Add(nC);
                        m_rows[current_row].m_data.Add(sum);
                    }
                    reader.Close();
                }
                stream.Close();
            }
            //end

            //compute norms
            foreach (Row row in m_rows)
            {
                row.norm = 0.0;
                foreach (short s in row.m_data)
                {
                    row.norm += s * s;
                }
                row.norm = Math.Sqrt(row.norm);
            }
            //end

            int NN = m_rows.Count - nCategories;
            for (int loop = 0; loop < NN; ++loop)
            {
                int nMatrixSize = m_rows.Count;
                float[,] fMatrix = new float[nMatrixSize, nMatrixSize];
                for (int i = 0; i < nMatrixSize; ++i)
                {
                    for (int j = i + 1; j < nMatrixSize; ++j)
                    {
                        fMatrix[i, j] = getCosine(m_rows[i], m_rows[j]);
                    }
                }

                int nrow = 0;
                int ncol = 0;
                float fMax = 0.0f;
                getMaxElement(ref nrow, ref ncol, ref fMax, fMatrix, nMatrixSize);

                Row r1 = m_rows[nrow];
                Row r2 = m_rows[ncol];
                Row merged = mergeTwoRows(r1, r2);
                m_rows.Remove(r1);
                m_rows.Remove(r2);
                m_rows.Add(merged);
 
                //count number of merged nodes
                int mergedNodes = 0;
                foreach (Row row in m_rows)
                {
                    if (row.m_files.Count > 1)
                    {
                        ++mergedNodes;
                    }
                }
                if (mergedNodes >= nCategories) break;
            }

            int cnt = 1;
            foreach (Row row in m_rows)
            {
                if (row.m_files.Count > 1)
                {
                    m_centroid.Add(row);
                }
                ++cnt;
            }
            return true;
        }

        private void showCentroids()
        {
            Console.WriteLine("Current centroids\n");
            foreach (Row row in m_centroid)
            {
                foreach (int n in row.m_files)
                {
                    Console.Write("{0,4}", n);
                }
                Console.WriteLine("\n");
            }
        }

        int getCategory(Row current)
        {
            double norm = 0.0;
            for (int k = 0; k < current.m_data.Count; ++k)
            {
                norm += current.m_data[k] * current.m_data[k];
            }
            norm = Math.Sqrt(norm);
            current.norm = norm;

            float max_cosine = 0.0f;
            int max_category = -1;
            for (int m = 0; m < m_centroid.Count; ++m)
            {
                float cosine = getCosine(current, m_centroid[m]);
                if (cosine > max_cosine)
                {
                    max_cosine = cosine;
                    max_category = m;
                }
            }
            return max_category;
        }

        private void assigningStep(string fileName)
        {
            m_categorizedFiles.Clear();

            FileInfo fi = new FileInfo(fileName);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);

            Row current = new Row();
            current.m_files.Add(0);

            int nR = 0;
            int nC = 0;
            short sum = 0;
            int cnt = 0;
            using (FileStream stream = new FileStream(fileName, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nRnew = reader.ReadInt32();
                        nC = reader.ReadInt32();
                        sum = reader.ReadInt16();

                        if (nRnew != nR)
                        {
                            int max_category = getCategory(current);
                            AssignedCategory acat = new AssignedCategory();
                            acat.m_category = max_category;
                            acat.m_fileNum = cnt;
                            m_categorizedFiles.Add(acat);
                            ++cnt;
                            
                            current.m_data.Clear();
                            current.m_pos.Clear();
                            current.m_files.Clear();
                            current.m_files.Add(nRnew);
                        }
                        nR = nRnew;

                        current.m_data.Add(sum);
                        current.m_pos.Add(nC);
                    }
                    reader.Close();
                }
                stream.Close();
            }

            int n = getCategory(current);
            AssignedCategory ac = new AssignedCategory();
            ac.m_category = n;
            ac.m_fileNum = cnt;
            m_categorizedFiles.Add(ac);
        }

        private void updatingStep(string fileName, int nCategories)
        {
            m_rows.Clear();
            int current_row = 0;
            m_rows.Add(new Row());
            m_rows[current_row].m_files.Add(current_row);

            FileInfo fi = new FileInfo(fileName);
            long nBytes = fi.Length;
            int nNonZeros = (int)(nBytes / 10);

            int nR = 0;
            int nC = 0;
            short sum = 0;
            using (FileStream stream = new FileStream(fileName, FileMode.Open))
            {
                using (BinaryReader reader = new BinaryReader(stream))
                {
                    for (long i = 0; i < nNonZeros; ++i)
                    {
                        int nRnew = reader.ReadInt32();
                        nC = reader.ReadInt32();
                        sum = reader.ReadInt16();

                        if (nRnew != nR)
                        {
                            m_rows.Add(new Row());
                            ++current_row;
                            m_rows[current_row].m_files.Add(current_row);
                        }
                        nR = nRnew;
                        m_rows[current_row].m_pos.Add(nC);
                        m_rows[current_row].m_data.Add(sum);
                    }
                    reader.Close();
                }
                stream.Close();
            }

            //compute norms
            foreach (Row row in m_rows)
            {
                row.norm = 0.0;
                foreach (short s in row.m_data)
                {
                    row.norm += s * s;
                }
                row.norm = Math.Sqrt(row.norm);
            }
            //end

            //foreach (Row row in m_rows)
            //{
            //    for (int i=0; i<row.m_files.Count; ++i) 
            //    {
            //        Console.WriteLine(" {0}  ", row.m_files[i]);
            //    }
            //}

            //for (int i = 0; i < m_categorizedFiles.Count; ++i)
            //{
            //    Console.WriteLine("{0} {1} ", m_categorizedFiles[i].m_fileNum, m_categorizedFiles[i].m_category);
            //}

            m_centroid.Clear();

            List<int>[] cats = new List<int>[nCategories];
            for (int i = 0; i < nCategories; ++i)
            {
                cats[i] = new List<int>(0);
            }
            for (int i = 0; i < nCategories; ++i)
            {
                for (int j = 0; j < m_categorizedFiles.Count; ++j)
                {
                    if (m_categorizedFiles[j].m_category == i)
                    {
                        cats[i].Add(m_categorizedFiles[j].m_fileNum);
                    }
                }
            }

            //for (int i = 0; i < nCategories; ++i)
            //{
            //    for (int j = 0; j < cats[i].Count; ++j)
            //    {
            //        Console.Write(" {0} ", cats[i][j]);
            //    }
            //    Console.WriteLine();
            //}

            for (int i = 0; i < nCategories; ++i)
            {
                if (cats[i].Count == 1)
                {
                    int index = cats[i][0];
                    m_centroid.Add(m_rows[index]);
                }
                else if (cats[i].Count == 2)
                {
                    int index1 = cats[i][0];
                    int index2 = cats[i][1];
                    Row row = mergeTwoRows(m_rows[index1], m_rows[index2]);
                    m_centroid.Add(row);
                }
                else if (cats[i].Count > 2)
                {
                    int index1 = cats[i][0];
                    int index2 = cats[i][1];
                    Row row = mergeTwoRows(m_rows[index1], m_rows[index2]);
                    for (int k = 2; k < cats[i].Count; ++k)
                    {
                        row = mergeTwoRows(row, m_rows[cats[i][k]]);
                    }
                    m_centroid.Add(row);
                }
            }
        }

        private void estimateAccuracyForPatentCorpus128()
        {
            int[,] nAccuracy = new int[8, 8];

            int cnt = 1;
            foreach (Row row in m_centroid)
            {
                for (int k = 0; k < 8; ++k)
                {
                    nAccuracy[cnt - 1, k] = 0;
                }
                foreach (int n in row.m_files)
                {
                    if (n < 16) ++nAccuracy[cnt - 1, 0];
                    else if (n < 32) ++nAccuracy[cnt - 1, 1];
                    else if (n < 48) ++nAccuracy[cnt - 1, 2];
                    else if (n < 64) ++nAccuracy[cnt - 1, 3];
                    else if (n < 80) ++nAccuracy[cnt - 1, 4];
                    else if (n < 96) ++nAccuracy[cnt - 1, 5];
                    else if (n < 112) ++nAccuracy[cnt - 1, 6];
                    else ++nAccuracy[cnt - 1, 7];
                }
                ++cnt;
            }

            for (int i = 0; i < 8; ++i)
            {
                for (int j = 0; j < 8; ++j)
                {
                    Console.Write("{0,4}", nAccuracy[i, j]);
                }
                Console.WriteLine();
            }

            int nCorrect = 0;
            for (int loop = 0; loop < 8; ++loop)
            {
                int RR = 0;
                int CC = 0;
                int max = nAccuracy[RR, CC];
                for (int i = 0; i < 8; ++i)
                {
                    for (int j = 0; j < 8; ++j)
                    {
                        if (nAccuracy[i, j] > max)
                        {
                            max = nAccuracy[i, j];
                            RR = i;
                            CC = j;
                        }
                    }
                }
                nCorrect += max;
                for (int i = 0; i < 8; ++i)
                {
                    nAccuracy[RR, i] = 0;
                }
                for (int j = 0; j < 8; ++j)
                {
                    nAccuracy[j, CC] = 0;
                }
            }
            Console.WriteLine("Correctly recognized {0} out of 128, ratio = {1:f}\n", nCorrect, (double)(nCorrect) / 128.0);
        }

        public bool computeKmeans(string fileName, int nCategories)
        {
            Console.WriteLine("Start k-means");
            if (!computeInitialCentroids(fileName, nCategories)) 
            {
                Console.WriteLine("Failed to initialize centroids");
                return false;
            }

            for (int i = 0; i < 5; ++i)
            {
                assigningStep(fileName);
                updatingStep(fileName, nCategories);
            }
  
            showCentroids();
            estimateAccuracyForPatentCorpus128();

            return true;
        }
    }
}